# requires ctypes 0.9.3 or later
import new
import warnings
import types
try:
    set
except NameError:
    from sets import Set as set
from ctypes import *

class IDLWarning(UserWarning):
    "Warn about questionable type information"

from comtypes.GUID import GUID
_GUID = GUID
IID = GUID
DWORD = c_ulong

################################################################
# constants for object creation
CLSCTX_INPROC_SERVER = 1
CLSCTX_INPROC_HANDLER = 2
CLSCTX_LOCAL_SERVER = 4

CLSCTX_INPROC = 3
CLSCTX_SERVER = 5
CLSCTX_ALL = 7

CLSCTX_INPROC_SERVER16 = 8
CLSCTX_REMOTE_SERVER = 16
CLSCTX_INPROC_HANDLER16 = 32
CLSCTX_RESERVED1 = 64
CLSCTX_RESERVED2 = 128
CLSCTX_RESERVED3 = 256
CLSCTX_RESERVED4 = 512
CLSCTX_NO_CODE_DOWNLOAD = 1024
CLSCTX_RESERVED5 = 2048
CLSCTX_NO_CUSTOM_MARSHAL = 4096
CLSCTX_ENABLE_CODE_DOWNLOAD = 8192
CLSCTX_NO_FAILURE_LOG = 16384
CLSCTX_DISABLE_AAA = 32768
CLSCTX_ENABLE_AAA = 65536
CLSCTX_FROM_DEFAULT_CONTEXT = 131072

tagCLSCTX = c_int # enum
CLSCTX = tagCLSCTX

################################################################
# Initialization and shutdown
_ole32 = oledll.ole32

_ole32.CoInitialize(None)

class _Cleaner(object):
    def __del__(self, func=_ole32.CoUninitialize):
        # Sometimes, CoUnititialize, running at Python shutdown, raises an exception.
        # We suppress this when __debug__ is False.
        if __debug__:
            func()
        else:
            try: func()
            except WindowsError: pass

__cleaner = _Cleaner()
del _Cleaner

def _clean_exc_info():
    # the purpose of this function is to ensure that no com object
    # pointers are in sys.exc_info()
    try: 1//0
    except: pass

import atexit
atexit.register(_clean_exc_info)

################################################################
# The metaclasses...

class _cominterface_meta(type):
    # Metaclass for COM interface classes.
    # Creates also a POINTER type for the newly created class.
    def __new__(self, name, bases, namespace):
        methods = namespace.pop("_methods_", None)
        dispmethods = namespace.pop("_disp_methods_", None)
        cls = type.__new__(self, name, bases, namespace)
        if methods is not None:
            setattr(cls, "_methods_", methods)
        if dispmethods is not None:
            setattr(cls, "_disp_methods_", dispmethods)

        # If we sublass a COM interface, for example:
        #
        # class IDispatch(IUnknown):
        #     ....
        #
        # then we need to make sure that POINTER(IDispatch) is a
        # subclass of POINTER(IUnknown) because of the way ctypes
        # typechecks work.
        if bases == (object,):
            _ptr_bases = (cls, _compointer_base)
        else:
            _ptr_bases = (cls, POINTER(bases[0]))
        # The interface 'cls' is used as a mixin.
        p = type(_compointer_base)("POINTER(%s)" % cls.__name__,
                                   _ptr_bases,
                                   {"__ctypes_from_outparam__": lambda self: self})
        from ctypes import _pointer_type_cache
        _pointer_type_cache[cls] = p
        return cls

    def __setattr__(self, name, value):
        if name == "_methods_":
            self._make_methods(value)
        elif name == "_disp_methods_":
            self._make_dispmethods(value)
        type.__setattr__(self, name, value)

    def _make_dispmethods(self, methods):
        # create dispinterface methods and properties on the interface 'self'
        for m in methods:
            what, name, idlflags, restype, argspec = m
            # argspec is a sequence of tuples, each tuple is:
            # ([paramflags], type, name)
            try:
                memid = [x for x in idlflags if isinstance(x, int)][0]
            except IndexError:
                raise TypeError, "no dispid found in idlflags"
            if what == "DISPPROPERTY": # DISPPROPERTY
                assert not argspec # XXX does not yet work for properties with parameters
                accessor = self._disp_property(memid, idlflags)
                setattr(self, name, accessor)
            elif what == "DISPMETHOD": # DISPMETHOD
                # method
                pass

    def _disp_property(self, memid, idlflags):
        def get(obj):
            return obj.Invoke(memid, _invkind=2) # DISPATCH_PROPERTYGET
        if "readonly" in idlflags:
            return property(get)
        def set(obj, value):
            return obj.Invoke(memid, value, _invkind=4) # DISPATCH_PROPERTYPUT
        return property(get, set)

    def __get_baseinterface_methodcount(self):
        "Return the number of com methods in the base interfaces"
        try:
            return sum([len(itf.__dict__["_methods_"])
                        for itf in self.mro()[1:-1]])
        except KeyError, (name,):
            if name == "_methods_":
                raise TypeError, "baseinterface '%s' has no _methods_" % itf.__name__
            raise

    def _make_methods(self, methods):
        # we insist on an _iid_ in THIS class!
        try:
            self.__dict__["_iid_"]
        except KeyError:
            raise AttributeError, "this class must define an _iid_"
        vtbl_offset = self.__get_baseinterface_methodcount()

        getters = {}
        setters = {}

        # create private low level, and public high level methods
        for i, item in enumerate(methods):
            restype, name, argtypes, paramflags, idlflags, doc = item
            # the function prototype
            prototype = WINFUNCTYPE(restype, *argtypes)

            # a low level unbound method calling the com method.
            # attach it with a private name (__com_AddRef, for example),
            # so that custom method implementations can call it.
            raw_func = prototype(i + vtbl_offset, name)
            setattr(self,
                    "_%s__com_%s" % (self.__name__, name),
                    new.instancemethod(raw_func, None, self))

            # a high level function calling the COM method
            func = prototype(i + vtbl_offset, name, paramflags)
            func.__doc__ = doc
            # make it an unbound method.  Remember, 'self' is a type here.
            mth = new.instancemethod(func, None, self)

            # is it a property set or property get?
            is_prop = False

            # XXX Hm.  What, when paramflags is None?
            # Or does have '0' values?
            # Seems we loose then, at least for properties...

            # The following code assumes that the docstrings for
            # propget and propput are identical.
            if "propget" in idlflags:
                assert name.startswith("_get_")
                nargs = len([flags for flags in paramflags
                             if flags[0] & 1])
                # XXX or should we do this?
                # nargs = len([flags for flags in paramflags
                #             if (flags[0] & 1) or (flags[0] == 0)])
                propname = name[len("_get_"):]
                getters[propname, doc, nargs] = func
                is_prop = True
            elif "propput" in idlflags:
                assert name.startswith("_set_")
                nargs = len([flags for flags in paramflags
                              if flags[0] & 1]) - 1
                propname = name[len("_set_"):]
                setters[propname, doc, nargs] = func
                is_prop = True

            # We install the method in the class, except when it's a
            # property accessor.  And we make sure we don't overwrite
            # a property that's already present in the class.
            if not is_prop:
                if hasattr(self, name):
                    setattr(self, "_" + name, mth)
                else:
                    setattr(self, name, mth)

        # create public properties / attribute accessors
        # XXX Das sieht mir aber komisch aus!
        for item in set(getters.keys()) | set(getters.keys()):
            name, doc, nargs = item
            if nargs == 0:
                prop = property(getters.get(item), setters.get(item), doc=doc)
            else:
                # Hm, must be a descriptor where the __get__ method
                # returns a bound object having __getitem__ and
                # __setitem__ methods.
                prop = named_property(getters.get(item), setters.get(item), doc=doc)
            # Again, we should not overwrite class attributes that are
            # already present.
            if hasattr(self, name):
                setattr(self, "_" + name, prop)
            else:
                setattr(self, name, prop)

################################################################
# helper classes for COM propget / propput
# Should they be implemented in C for speed?

class bound_named_property(object):
    def __init__(self, getter, setter, im_inst):
        self.im_inst = im_inst
        self.getter = getter
        self.setter = setter

    def __getitem__(self, index):
        if self.getter is None:
            raise TypeError("unsubscriptable object")
        return self.getter(self.im_inst, index)

    def __setitem__(self, index, value):
        if self.setter is None:
            raise TypeError("object does not support item assignment")
        self.setter(self.im_inst, index, value)

class named_property(object):
    def __init__(self, getter, setter, doc):
        self.getter = getter
        self.setter = setter
        self.doc = doc

    def __get__(self, im_inst, im_class=None):
        if im_inst is None:
            return self
        return bound_named_property(self.getter, self.setter, im_inst)

################################################################

class _compointer_meta(type(c_void_p), _cominterface_meta):
    "metaclass for COM interface pointer classes"
    # (no functionality, but avoids a metaclass confict)

class _compointer_base(c_void_p):
    "base class for COM interface pointer classes"
    __metaclass__ = _compointer_meta
    def __del__(self):
        "Release the COM refcount we own."
        if self: # calls __nonzero__
            self.Release()

    # Hm, shouldn't this be in c_void_p ?
    def __nonzero__(self):
        # get the .value property of the baseclass, this is the pointer value
        return bool(super(_compointer_base, self).value)

    def __cmp__(self, other):
        # COM identity rule
        #
        # XXX To compare COM interface pointers, should we
        # automatically QueryInterface for IUnknown on both items, and
        # compare the pointer values?
        if not isinstance(other, _compointer_base):
            return 1

        # get the value property of the c_void_p baseclass, this is the pointer value
        return cmp(super(_compointer_base, self).value, super(_compointer_base, other).value)

    # override the .value property of c_void_p
    #
    # for symmetry with other ctypes types
    # XXX explain
    # XXX check if really needed
    def __get_value(self):
        return self
    value = property(__get_value)

    def __repr__(self):
        return "<%s instance at %x>" % (self.__class__.__name__, id(self))

################################################################

from ctypes import _SimpleCData

class BSTR(_SimpleCData):
    "The windows BSTR data type"
    _type_ = "X"
    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.value)
    def __del__(self, _free=windll.oleaut32.SysFreeString):
        if not self._b_base_:
            _free(self)

################################################################
# IDL stuff

class helpstring(object):
    "Specifies the helpstring for a COM method or property."
    def __init__(self, text):
        self.text = text

class defaultvalue(object):
    "Specifies the default value for parameters marked optional."
    def __init__(self, value):
        self.value = value

class dispid(int):
    "Specifies the DISPID of a method or property."
    pass

def STDMETHOD(restype, name, argtypes=()):
    "Specifies a COM method slot without idlflags"
    # restype, name, argtypes, paramflags, idlflags, docstring
    return restype, name, argtypes, None, (), None

def DISPMETHOD(idlflags, restype, name, *argspec):
    "Specifies a method of a dispinterface"
    return "DISPMETHOD", name, idlflags, restype, argspec

def DISPPROPERTY(idlflags, proptype, name):
    "Specifies a property of a dispinterface"
    return "DISPPROPERTY", name, idlflags, proptype, ()#, argspec

# COMMETHOD returns:
# restype, methodname, tuple(argtypes), tuple(paramflags), tuple(idlflags), helptext
#
# paramflags is a sequence of (flags (integer), paramname (string)
# tuple(idlflags) is for the method itself: (dispid, 'readonly')
#
# Example: (HRESULT, 'Width', (c_long,), (2, 'rhs'), (4, 'readonly'), None)

## sample generated code:
##    DISPPROPERTY([5, 'readonly'], OLE_YSIZE_HIMETRIC, 'Height'),
##    DISPMETHOD([6], None, 'Render',
##               ( [], c_int, 'hdc' ),
##               ( [], c_int, 'x' ),
##               ( [], c_int, 'y' ))

################################################################

_PARAMFLAGS = {
    "in": 1,
    "out": 2,
    "retval": 8,
    "optional": 16,
    # PARAMFLAG_HASDEFAULT: 0x20
    }

def _encode_idl(names):
    # sum up all values found in _PARAMFLAGS, ignoring all others.
    result = sum([_PARAMFLAGS.get(n, 0) for n in names])
    return result & 3 # that's what _ctypes accept

_NOTHING = object()
def _unpack_argspec(idl, typ, name=None, defval=_NOTHING):
    return idl, typ, name, defval

# will be overwritten with VARIANT when comtypes.automation is imported.
_VARIANT_type_hack = None

def COMMETHOD(idlflags, restype, methodname, *argspec):
    """Specifies a COM method slot with idlflags.

    XXX more
    """
    paramflags = []
    argtypes = []

    # collect all helpstring instances
    helptext = [t.text for t in idlflags if isinstance(t, helpstring)]
    # join them together(does this make sense?) and replace by None if empty.
    helptext = "".join(helptext) or None

    for item in argspec:
        idl, typ, argname, defval = _unpack_argspec(*item)
        pflags = _encode_idl(idl)
        if "optional" in idl:
            if defval is _NOTHING:
                # XXX We don't want to make this module import
                # comtypes.automation but we need VARIANT.  So
                # comtypes.automation sets it from the outside into
                # this module, when it is imported. Hack, hack.
                if _VARIANT_type_hack and typ is _VARIANT_type_hack:
                    # XXX make this a classmethod of VARIANT!
                    # VARIANT.missing or so...
                    v = _VARIANT_type_hack()
                    VT_ERROR = 10
                    VT_EMPTY = 0
                    v.vt = VT_EMPTY
##                    v.vt = VT_ERROR
##                    v._.VT_I4 = 0x80020004L
                    defval = v
##                    defval = _VARIANT_type_hack()
                elif _VARIANT_type_hack and typ is POINTER(_VARIANT_type_hack):
                    v = _VARIANT_type_hack()
                    VT_ERROR = 10
##                    VT_EMPTY = 0
##                    v.vt = VT_EMPTY
                    v.vt = VT_ERROR
                    v._.VT_I4 = 0x80020004L

                    defval = pointer(v)
                else:
##                    msg = "'optional' only allowed for VARIANT and VARIANT*, not for %s" \
##                                  % typ.__name__
##                    warnings.warn(msg, IDLWarning, stacklevel=2)
                    defval = typ()
            paramflags.append((pflags, argname, defval))
        else:
            paramflags.append((pflags, argname))
        argtypes.append(typ)
    if "propget" in idlflags:
        methodname = "_get_%s" % methodname
    elif "propput" in idlflags:
        methodname = "_set_%s" % methodname
    return restype, methodname, tuple(argtypes), tuple(paramflags), tuple(idlflags), helptext

################################################################
# IUnknown, the root of all evil...

class IUnknown(object):
    """The most basic COM interface.

    Each subclasses of IUnknown must define these class attributes:

    _iid_ - a GUID instance defining the identifier of this interface

    _methods_ - a list of methods for this interface.

    The _methods_ list must in VTable order.  Methods are specified
    with STDMETHOD or COMMETHOD calls.
    """
    __metaclass__ = _cominterface_meta
    _iid_ = GUID("{00000000-0000-0000-C000-000000000046}")

    _methods_ = [
        STDMETHOD(HRESULT, "QueryInterface",
                  [POINTER(GUID), POINTER(c_void_p)]),
        STDMETHOD(c_ulong, "AddRef"),
        STDMETHOD(c_ulong, "Release")
    ]

    def QueryInterface(self, interface, iid=None):
        "QueryInterface(klass) -> instance"
        p = POINTER(interface)()
        if iid is None:
            iid = interface._iid_
        self.__com_QueryInterface(byref(iid), byref(p))
        return p

    # these are only so that they get a docstring.
    # XXX There should be other ways to install a docstring.
    def AddRef(self):
        "Increase the internal refcount by one"
        return self.__com_AddRef()

    def Release(self):
        "Decrease the internal refcount by one"
        return self.__com_Release()

    # should these methods be in a mixin class, which the metaclass
    # adds when it detects the Count, Item, and _NewEnum methods?
    #
    # Hm, codegenerator adds __iter__, next, and __getitem__ when
    # it detects an IEnumXXX interface...
    def __len__(self):
        try:
            length = self.Count
        except AttributeError:
            raise TypeError, "len() of unsized object"
        else:
            return length

    # does this make sense? It seems that all standard typelibs I've
    # seen so far that support .Item aso support ._NewEnum
    #
    # And how can we know if the indexes are 1-based or 0-based?
    
    def __getitem__(self, index):
        # Should we insist on a Count method also?
        # And/Or we should insits that the Item method has a dispid of DISPID_VALUE ( 0 ).
        try:
            mth = self.Item
        except AttributeError:
            raise TypeError, "unsubscriptable object"
        try:
            result = mth(index + 1)
        except WindowsError, detail:
            raise IndexError, "invalid index"
        if not result: # we got a NULL com pointer
            raise IndexError, "invalid index"
        # Hm, _NewEnum returns a IEnumVARIANT pointer.
        # This has items of IUnknown*.  What does .Item() have? What about 'wrap'?
        return result

    def __make_iter(self):
        enum = self._NewEnum
        if isinstance(enum, types.MethodType):
            # _NewEnum should be a propget property, with dispid -4.  See:
            # http://msdn.microsoft.com/library/en-us/automat/htm/chap2_2ws9.asp
            # http://msdn.microsoft.com/library/en-us/automat/htm/chap4_64j7.asp
            #
            # Sometimes, however, it is a method.
            enum = enum()
        # _NewEnum returns an IUnknown pointer,
        # QueryInterface() it to IEnumVARIANT
        from comtypes.automation import IEnumVARIANT
        return enum.QueryInterface(IEnumVARIANT)

    # Some magic to implement an __iter__ method.  Raises
    # attributeerror if no _NewEnum attribute is found in the class.
    # If _NewEnum is present, returns a callable that will return a
    # python iterator when called.  Thanks to Bengt Richter for the
    # idea.
    def __getIter(self):
        try:
            enum = self._NewEnum
        except AttributeError:
            raise AttributeError, "__iter__"
        return self.__make_iter
    __iter__ = property(__getIter)

################################################################

def CoCreateInstance(clsid, interface=IUnknown, clsctx=CLSCTX_ALL, punkouter=None):
    """The basic windows api to create a COM class object and return a
    pointer to an interface.
    """
    p = POINTER(interface)()
    iid = interface._iid_
    CoCreateInstance._api_(byref(clsid), punkouter, clsctx, byref(iid), byref(p))
    return p
CoCreateInstance = stdcall(HRESULT, 'ole32',
                           [POINTER(IID), POINTER(IUnknown), c_ulong,
                            POINTER(IID), POINTER(c_void_p)]) (CoCreateInstance)

################################################################
# What's a coclass?
# a POINTER to a coclass is allowed as parameter in a function declaration:
# http://msdn.microsoft.com/library/en-us/midl/midl/oleautomation.asp
class CoClass(object):
    from comtypes._meta import _coclass_meta as __metaclass__

################
##class IErrorInfo(IUnknown):
##    _iid_ = GUID("{1CF2B120-547D-101B-8E65-08002B2BD119}")
##    _methods_ = [
##        # C:/Programme/gccxml/bin/Vc71/PlatformSDK/oaidl.h 5186
##        STDMETHOD(HRESULT, 'GetGUID', [POINTER(GUID)]),
##        STDMETHOD(HRESULT, 'GetSource', [POINTER(BSTR)]),
##        STDMETHOD(HRESULT, 'GetDescription', [POINTER(BSTR)]),
##        STDMETHOD(HRESULT, 'GetHelpFile', [POINTER(BSTR)]),
##        STDMETHOD(HRESULT, 'GetHelpContext', [POINTER(DWORD)]),
##    ]

################
##class ISupportErrorInfo(IUnknown):
##    _iid_ = GUID("{DF0B3D60-548F-101B-8E65-08002B2BD119}")
##    _methods_ = [
##        # C:/Programme/gccxml/bin/Vc71/PlatformSDK/oaidl.h 5546
##        STDMETHOD(HRESULT, 'InterfaceSupportsErrorInfo', [POINTER(IID)]),
##    ]

##__all__ = ["CoClass", "IUnknown", "GUID", "HRESULT", "BSTR", "STDMETHOD", "COMMETHOD"]
